"""Team resolution logic for chains."""

from typing import Callable, Optional

import click

from truss.cli import remote_cli
from truss.remote.baseten.custom_types import TeamType
from truss.remote.baseten.remote import BasetenRemote


def _validate_provided_team(
    provided_team_name: str, existing_teams: dict[str, TeamType]
) -> str:
    """Validate provided team name exists and return team_id."""
    if provided_team_name not in existing_teams:
        available_teams_str = remote_cli.format_available_teams(existing_teams)
        raise click.ClickException(
            f"Team '{provided_team_name}' does not exist. Available teams: {available_teams_str}"
        )
    return existing_teams[provided_team_name].id


def _get_matching_chains(
    chain_name: str,
    existing_teams: dict[str, TeamType],
    fetch_chains: Callable[[], list],
) -> list[dict]:
    """Get chains matching name that are in accessible teams."""
    all_chains = fetch_chains()
    accessible_team_ids = {team_data.id for team_data in existing_teams.values()}
    return [
        c
        for c in all_chains
        if c.get("name") == chain_name
        and c.get("team", {}).get("id") in accessible_team_ids
    ]


def _prompt_for_team_from_chains(
    matching_chains: list[dict], existing_teams: dict[str, TeamType], prompt: str
) -> dict:
    """Prompt user to select team when multiple chains match, return selected chain."""
    team_name_to_chain = {
        c.get("team", {}).get("name", "Unknown"): c for c in matching_chains
    }
    teams_with_chain = {
        name: existing_teams[name]
        for name in team_name_to_chain
        if name in existing_teams
    }
    if not teams_with_chain:
        raise click.ClickException(
            "Chain exists but you don't have access to any team that owns it."
        )
    selected = remote_cli.inquire_team(existing_teams=teams_with_chain, prompt=prompt)
    if not selected or selected not in team_name_to_chain:
        raise click.ClickException("No team selected.")
    return team_name_to_chain[selected]


def resolve_chain_for_watch(
    remote_provider: BasetenRemote,
    chain_name: str,
    provided_team_name: Optional[str] = None,
    prompt: str = "Multiple chains with this name exist. Which team's chain do you want to watch?",
) -> dict:
    """Resolve a chain by name for watch, handling team disambiguation.

    Returns chain dict with 'id', 'name', and 'team' info.
    """
    existing_teams = remote_provider.api.get_teams()

    if provided_team_name is not None:
        team_id = _validate_provided_team(provided_team_name, existing_teams)
        chains = remote_provider.api.get_chains(team_id=team_id)
        matching = [c for c in chains if c.get("name") == chain_name]
        if not matching:
            raise click.ClickException(
                f"Chain '{chain_name}' not found in team '{provided_team_name}'."
            )
        return matching[0]

    matching_chains = _get_matching_chains(
        chain_name, existing_teams, remote_provider.api.get_chains
    )

    if not matching_chains:
        raise click.ClickException(f"Chain '{chain_name}' not found.")

    if len(matching_chains) == 1:
        return matching_chains[0]

    return _prompt_for_team_from_chains(matching_chains, existing_teams, prompt)


def resolve_chain_team_name(
    remote_provider: BasetenRemote,
    provided_team_name: Optional[str],
    existing_chain_name: Optional[str] = None,
    existing_teams: Optional[dict[str, TeamType]] = None,
) -> tuple[Optional[str], Optional[str]]:
    """Resolve team name and team_id from provided team name or by prompting the user.
    Returns a tuple of (team_name, team_id).
    This function handles 8 distinct scenarios organized into 3 high-level categories:

    HIGH-LEVEL SCENARIO 1: --team PROVIDED
        SCENARIO 1: Valid team name, user has access
            → Returns (team_name, team_id) for that team (no prompt, no error)
        SCENARIO 2: Invalid team name (does not exist)
            → Raises ClickException with error message listing available teams

    HIGH-LEVEL SCENARIO 2: --team NOT PROVIDED, Chain does not exist
        SCENARIO 3: User has multiple teams, no existing chain
            → Prompts user to select a team via inquire_team()
        SCENARIO 6: User has exactly one team, no existing chain
            → Returns (team_name, team_id) for the single team automatically (no prompt)

    HIGH-LEVEL SCENARIO 3: --team NOT PROVIDED, Chain exists
        SCENARIO 4: User has multiple teams, existing chain in exactly one team
            → Auto-detects and returns (team_name, team_id) for that team (no prompt)
        SCENARIO 5: User has multiple teams, existing chain exists in multiple teams
            → Prompts user to select a team via inquire_team()
        SCENARIO 7: User has exactly one team, existing chain matches the team
            → Auto-detects and returns (team_name, team_id) for the single team (no prompt)
        SCENARIO 8: User has exactly one team, existing chain exists in different team
            → Returns (team_name, team_id) for the single team automatically (no prompt, uses user's only team)
    """
    if existing_teams is None:
        existing_teams = remote_provider.api.get_teams()

    def _get_team_id(team_name: Optional[str]) -> Optional[str]:
        if team_name and existing_teams:
            team_data = existing_teams.get(team_name)
            return team_data.id if team_data else None
        return None

    if provided_team_name is not None:
        if provided_team_name not in existing_teams:
            available_teams_str = remote_cli.format_available_teams(existing_teams)
            raise click.ClickException(
                f"Team '{provided_team_name}' does not exist. Available teams: {available_teams_str}"
            )
        return (provided_team_name, _get_team_id(provided_team_name))

    existing_chains = None
    if existing_chain_name is not None:
        existing_chains = remote_provider.api.get_chains()
        matching_chains = [
            c for c in existing_chains if c.get("name") == existing_chain_name
        ]

        if len(matching_chains) > 1:
            selected_team_name = remote_cli.inquire_team(existing_teams=existing_teams)
            return (selected_team_name, _get_team_id(selected_team_name))

        if len(matching_chains) == 1:
            chain_team = matching_chains[0].get("team")
            chain_team_name = chain_team.get("name") if chain_team else None
            if chain_team_name and chain_team_name in existing_teams:
                return (chain_team_name, _get_team_id(chain_team_name))

    if len(existing_teams) == 1:
        single_team_name = list(existing_teams.keys())[0]
        return (single_team_name, _get_team_id(single_team_name))

    selected_team_name = remote_cli.inquire_team(existing_teams=existing_teams)
    return (selected_team_name, _get_team_id(selected_team_name))



"""Chain gatherer for bundling external packages into a chain archive.

This module provides functionality to gather external_package_dirs from chainlets
into a self-contained chain archive, similar to how regular Truss handles
external packages via truss_handle/truss_gatherer.py.
"""

import pathlib
import tempfile

from truss.truss_handle.truss_handle import TrussHandle
from truss.util.path import copy_file_path, copy_tree_path

BUNDLED_PACKAGES_DIR = "packages"


def gather_chain(
    chain_root: pathlib.Path, external_package_dirs: list[pathlib.Path]
) -> pathlib.Path:
    """
    Creates a gathered version of the chain that includes all external_package_dirs.

    Args:
        chain_root: The root directory of the chain.
        external_package_dirs: List of external package directories to bundle.
            Caller is responsible for ensuring this list is non-empty and deduplicated.

    Returns:
        A new temp directory containing the chain with bundled packages.
    """
    # Create gathered chain in temp directory
    gathered_chain_root = pathlib.Path(tempfile.mkdtemp(prefix="gathered_chain_"))

    # Copy chain_root contents to gathered directory
    copy_tree_path(chain_root, gathered_chain_root)

    # Create packages directory for bundled external packages
    packages_dir = gathered_chain_root / BUNDLED_PACKAGES_DIR
    packages_dir.mkdir(exist_ok=True)

    for ext_dir in external_package_dirs:
        if not ext_dir.is_dir():
            raise ValueError(
                f"External packages directory at {ext_dir} is not a directory"
            )

        # Copy contents of the external package directory, not the directory itself.
        # This mimics the behavior in truss_gatherer.py and replicates adding
        # external package directory to sys.path.
        for sub_path in ext_dir.iterdir():
            dest_path = packages_dir / sub_path.name
            if sub_path.is_dir():
                copy_tree_path(sub_path, dest_path)
            elif sub_path.is_file():
                copy_file_path(sub_path, dest_path)

    # Clear external_package_dirs from all chainlet configs so the gathered chain
    # can be re-pushed without needing the original external paths.
    # This mirrors truss_gatherer.py's shadow_handle.clear_external_packages().
    _clear_external_package_dirs_from_configs(gathered_chain_root)

    return gathered_chain_root


def _clear_external_package_dirs_from_configs(chain_root: pathlib.Path) -> None:
    """Clear external_package_dirs from all chainlet config.yaml files.

    After gathering, external packages are bundled into packages/, so the
    original external_package_dirs paths are no longer needed. Clearing them
    allows the gathered chain to be downloaded and re-pushed without errors.

    Reuses TrussHandle.clear_external_packages() since each chainlet directory
    is a valid truss directory structure.
    """
    for chainlet_dir in chain_root.glob("chainlet_*"):
        if chainlet_dir.is_dir():
            handle = TrussHandle(chainlet_dir, validate=False)
            if handle.spec.config.external_package_dirs:
                handle.clear_external_packages()



import pathlib

CUSTOM = "custom"

_TRUSS_ROOT = pathlib.Path(__file__).parent.parent.resolve()

TEMPLATES_DIR = _TRUSS_ROOT / "templates"
TRADITIONAL_CUSTOM_TEMPLATE_DIR = TEMPLATES_DIR / "custom"
PYTHON_DX_CUSTOM_TEMPLATE_DIR = TEMPLATES_DIR / "custom_python_dx"
DOCKER_SERVER_TEMPLATES_DIR = TEMPLATES_DIR / "docker_server"
SERVER_CODE_DIR: pathlib.Path = TEMPLATES_DIR / "server"
TRITON_SERVER_CODE_DIR: pathlib.Path = TEMPLATES_DIR / "triton"
TRTLLM_TRUSS_DIR: pathlib.Path = TEMPLATES_DIR / "trtllm-briton"
SHARED_SERVING_AND_TRAINING_CODE_DIR_NAME = "shared"
SHARED_SERVING_AND_TRAINING_CODE_DIR: pathlib.Path = (
    TEMPLATES_DIR / SHARED_SERVING_AND_TRAINING_CODE_DIR_NAME
)
CONTROL_SERVER_CODE_DIR: pathlib.Path = TEMPLATES_DIR / "control"
CHAINS_CODE_DIR: pathlib.Path = _TRUSS_ROOT.parent / "truss-chains" / "truss_chains"
TRUSS_CODE_DIR: pathlib.Path = _TRUSS_ROOT.parent / "truss"
TRAINING_TEMPLATE_DIR = TEMPLATES_DIR / "train"
# Must be sorted ascendingly.
SUPPORTED_PYTHON_VERSIONS = ["3.9", "3.10", "3.11", "3.12", "3.13", "3.14"]

TRTLLM_PREDICT_CONCURRENCY = 512
BEI_TRTLLM_CLIENT_BATCH_SIZE = 128
BEI_MAX_CONCURRENCY_TARGET_REQUESTS = 2048
BEI_REQUIRED_MAX_NUM_TOKENS = 16384

TRTLLM_MIN_MEMORY_REQUEST_GI = 10
HF_MODELS_API_URL = "https://huggingface.co/api/models"
HF_ACCESS_TOKEN_KEY = "hf_access_token"
TRUSSLESS_MAX_PAYLOAD_SIZE = "64M"
# Alias for TEMPLATES_DIR
SERVING_DIR: pathlib.Path = TEMPLATES_DIR

REQUIREMENTS_TXT_FILENAME = "requirements.txt"
USER_SUPPLIED_REQUIREMENTS_TXT_FILENAME = "user_requirements.txt"
BASE_SERVER_REQUIREMENTS_TXT_FILENAME = "base_server_requirements.txt"
SERVER_REQUIREMENTS_TXT_FILENAME = "server_requirements.txt"
SYSTEM_PACKAGES_TXT_FILENAME = "system_packages.txt"

FILENAME_CONSTANTS_MAP = {
    "config_requirements_filename": REQUIREMENTS_TXT_FILENAME,
    "user_supplied_requirements_filename": USER_SUPPLIED_REQUIREMENTS_TXT_FILENAME,
    "base_server_requirements_filename": BASE_SERVER_REQUIREMENTS_TXT_FILENAME,
    "server_requirements_filename": SERVER_REQUIREMENTS_TXT_FILENAME,
    "system_packages_filename": SYSTEM_PACKAGES_TXT_FILENAME,
}

SERVER_DOCKERFILE_TEMPLATE_NAME = "server.Dockerfile.jinja"
NO_BUILD_DOCKERFILE_TEMPLATE_NAME = "no_build.Dockerfile.jinja"
MODEL_DOCKERFILE_NAME = "Dockerfile"
MODEL_CACHE_PATH = pathlib.Path("/app/model_cache")
README_TEMPLATE_NAME = "README.md.jinja"
MODEL_README_NAME = "README.md"

CONFIG_FILE = "config.yaml"
DOCKERFILE = "Dockerfile"
# Used to indicate whether to associate a container with Truss
TRUSS = "truss"
# Used to create unique identifier based on last time truss was updated
TRUSS_MODIFIED_TIME = "truss_modified_time"
# Path of the Truss used to identify which Truss is being referred
TRUSS_DIR = "truss_dir"
TRUSS_HASH = "truss_hash"


INFERENCE_SERVER_PORT = 8080
DEFAULT_NON_ROOT_USER_ID = 60000

HTTP_PUBLIC_BLOB_BACKEND = "http_public"

REGISTRY_BUILD_SECRET_PREFIX = "DOCKER_REGISTRY_"

# TODO: build the image so that the default path `python3` can be used - then remove here.
TRTLLM_PYTHON_EXECUTABLE = "/usr/local/briton/venv/bin/python"
# TODO: build the image so that the default path `python3` can be used - then remove here.
OPENAI_COMPATIBLE_TAG = "openai-compatible"
OPENAI_NON_COMPATIBLE_TAG = "force-legacy-api-non-openai-compatible"  # deprecated - openai-compatible is now the default


PRODUCTION_ENVIRONMENT_NAME = "production"

TRUSS_BASE_IMAGE_NAME = "baseten/truss-server-base"

DEFAULT_TRAINING_CHECKPOINT_FOLDER = "/tmp/loaded_checkpoints"



"""Tests for truss chains CLI commands."""

from unittest.mock import Mock, patch

from click.testing import CliRunner

from truss.cli.cli import truss_cli


def test_chains_push_with_disable_chain_download_flag():
    """Test that --disable-chain-download flag is properly parsed and passed through."""
    runner = CliRunner()

    mock_entrypoint_cls = Mock()
    mock_entrypoint_cls.meta_data.chain_name = "test_chain"
    mock_entrypoint_cls.display_name = "TestChain"

    mock_service = Mock()
    mock_service.run_remote_url = "http://test.com/run_remote"
    mock_service.is_websocket = False

    with patch(
        "truss_chains.framework.ChainletImporter.import_target"
    ) as mock_importer:
        with patch("truss_chains.deployment.deployment_client.push") as mock_push:
            mock_importer.return_value.__enter__.return_value = mock_entrypoint_cls
            mock_push.return_value = mock_service

            result = runner.invoke(
                truss_cli,
                [
                    "chains",
                    "push",
                    "test_chain.py",
                    "--disable-chain-download",
                    "--remote",
                    "test_remote",
                    "--dryrun",
                ],
            )

    assert result.exit_code == 0

    mock_push.assert_called_once()
    call_args = mock_push.call_args
    options = call_args[0][1]

    assert hasattr(options, "disable_chain_download")
    assert options.disable_chain_download is True


def test_chains_push_without_disable_chain_download_flag():
    """Test that disable_chain_download defaults to False when flag is not provided."""
    runner = CliRunner()

    mock_entrypoint_cls = Mock()
    mock_entrypoint_cls.meta_data.chain_name = "test_chain"
    mock_entrypoint_cls.display_name = "TestChain"

    mock_service = Mock()
    mock_service.run_remote_url = "http://test.com/run_remote"
    mock_service.is_websocket = False

    with patch(
        "truss_chains.framework.ChainletImporter.import_target"
    ) as mock_importer:
        with patch("truss_chains.deployment.deployment_client.push") as mock_push:
            mock_importer.return_value.__enter__.return_value = mock_entrypoint_cls
            mock_push.return_value = mock_service

            result = runner.invoke(
                truss_cli,
                [
                    "chains",
                    "push",
                    "test_chain.py",
                    "--remote",
                    "test_remote",
                    "--dryrun",
                ],
            )

    assert result.exit_code == 0

    mock_push.assert_called_once()
    call_args = mock_push.call_args
    options = call_args[0][1]

    assert hasattr(options, "disable_chain_download")
    assert options.disable_chain_download is False


def test_chains_push_help_includes_disable_chain_download():
    """Test that --disable-chain-download appears in the help output."""
    runner = CliRunner()

    result = runner.invoke(truss_cli, ["chains", "push", "--help"])

    assert result.exit_code == 0
    assert "--disable-chain-download" in result.output


def test_chains_push_with_deployment_name_flag():
    """Test that --deployment-name flag is properly parsed and passed through."""
    runner = CliRunner()

    mock_entrypoint_cls = Mock()
    mock_entrypoint_cls.meta_data.chain_name = "test_chain"
    mock_entrypoint_cls.display_name = "TestChain"

    mock_service = Mock()
    mock_service.run_remote_url = "http://test.com/run_remote"
    mock_service.is_websocket = False

    with patch(
        "truss_chains.framework.ChainletImporter.import_target"
    ) as mock_importer:
        with patch("truss_chains.deployment.deployment_client.push") as mock_push:
            mock_importer.return_value.__enter__.return_value = mock_entrypoint_cls
            mock_push.return_value = mock_service

            result = runner.invoke(
                truss_cli,
                [
                    "chains",
                    "push",
                    "test_chain.py",
                    "--deployment-name",
                    "custom_deployment",
                    "--remote",
                    "test_remote",
                    "--dryrun",
                ],
            )

    assert result.exit_code == 0

    mock_push.assert_called_once()
    call_args = mock_push.call_args
    options = call_args[0][1]

    assert hasattr(options, "deployment_name")
    assert options.deployment_name == "custom_deployment"


def test_chains_push_defaults_to_published():
    """Test that chains push defaults to published deployment."""
    runner = CliRunner()

    mock_entrypoint_cls = Mock()
    mock_entrypoint_cls.meta_data.chain_name = "test_chain"
    mock_entrypoint_cls.display_name = "TestChain"

    mock_service = Mock()
    mock_service.run_remote_url = "http://test.com/run_remote"
    mock_service.is_websocket = False

    with patch(
        "truss_chains.framework.ChainletImporter.import_target"
    ) as mock_importer:
        with patch("truss_chains.deployment.deployment_client.push") as mock_push:
            mock_importer.return_value.__enter__.return_value = mock_entrypoint_cls
            mock_push.return_value = mock_service

            result = runner.invoke(
                truss_cli,
                [
                    "chains",
                    "push",
                    "test_chain.py",
                    "--remote",
                    "test_remote",
                    "--dryrun",
                ],
            )

    assert result.exit_code == 0

    mock_push.assert_called_once()
    call_args = mock_push.call_args
    options = call_args[0][1]

    # Verify publish defaults to True (published deployment)
    assert options.publish is True


def test_chains_push_publish_flag_shows_deprecation_warning():
    """Test that --publish flag shows deprecation warning."""
    runner = CliRunner()

    mock_entrypoint_cls = Mock()
    mock_entrypoint_cls.meta_data.chain_name = "test_chain"
    mock_entrypoint_cls.display_name = "TestChain"

    mock_service = Mock()
    mock_service.run_remote_url = "http://test.com/run_remote"
    mock_service.is_websocket = False

    with patch(
        "truss_chains.framework.ChainletImporter.import_target"
    ) as mock_importer:
        with patch("truss_chains.deployment.deployment_client.push") as mock_push:
            mock_importer.return_value.__enter__.return_value = mock_entrypoint_cls
            mock_push.return_value = mock_service

            result = runner.invoke(
                truss_cli,
                [
                    "chains",
                    "push",
                    "test_chain.py",
                    "--publish",
                    "--remote",
                    "test_remote",
                    "--dryrun",
                ],
            )

    assert result.exit_code == 0
    assert "DEPRECATED" in result.output


def test_chains_push_no_publish_flag_shows_deprecation_warning():
    """Test that --no-publish flag shows deprecation warning but still works."""
    runner = CliRunner()

    mock_entrypoint_cls = Mock()
    mock_entrypoint_cls.meta_data.chain_name = "test_chain"
    mock_entrypoint_cls.display_name = "TestChain"

    mock_service = Mock()
    mock_service.run_remote_url = "http://test.com/run_remote"
    mock_service.is_websocket = False

    with patch(
        "truss_chains.framework.ChainletImporter.import_target"
    ) as mock_importer:
        with patch("truss_chains.deployment.deployment_client.push") as mock_push:
            mock_importer.return_value.__enter__.return_value = mock_entrypoint_cls
            mock_push.return_value = mock_service

            result = runner.invoke(
                truss_cli,
                [
                    "chains",
                    "push",
                    "test_chain.py",
                    "--no-publish",
                    "--remote",
                    "test_remote",
                    "--dryrun",
                ],
            )

    assert result.exit_code == 0
    assert "DEPRECATED" in result.output

    # Verify it still creates development deployment for backwards compatibility
    mock_push.assert_called_once()
    call_args = mock_push.call_args
    options = call_args[0][1]
    assert options.publish is False


def test_chains_push_watch_creates_development_deployment():
    """Test that --watch creates a development deployment."""
    runner = CliRunner()

    mock_entrypoint_cls = Mock()
    mock_entrypoint_cls.meta_data.chain_name = "test_chain"
    mock_entrypoint_cls.display_name = "TestChain"

    mock_service = Mock()
    mock_service.run_remote_url = "http://test.com/run_remote"
    mock_service.is_websocket = False

    with patch(
        "truss_chains.framework.ChainletImporter.import_target"
    ) as mock_importer:
        with patch("truss_chains.deployment.deployment_client.push") as mock_push:
            mock_importer.return_value.__enter__.return_value = mock_entrypoint_cls
            mock_push.return_value = mock_service

            result = runner.invoke(
                truss_cli,
                [
                    "chains",
                    "push",
                    "test_chain.py",
                    "--watch",
                    "--remote",
                    "test_remote",
                    "--dryrun",
                ],
            )

    assert result.exit_code == 0

    mock_push.assert_called_once()
    call_args = mock_push.call_args
    options = call_args[0][1]

    # Verify --watch creates development deployment
    assert options.publish is False


def test_chains_push_watch_with_promote_fails():
    """Test that --watch with --promote fails."""
    runner = CliRunner()

    mock_entrypoint_cls = Mock()
    mock_entrypoint_cls.meta_data.chain_name = "test_chain"
    mock_entrypoint_cls.display_name = "TestChain"

    with patch(
        "truss_chains.framework.ChainletImporter.import_target"
    ) as mock_importer:
        mock_importer.return_value.__enter__.return_value = mock_entrypoint_cls

        result = runner.invoke(
            truss_cli,
            [
                "chains",
                "push",
                "test_chain.py",
                "--watch",
                "--promote",
                "--remote",
                "test_remote",
            ],
        )

    assert result.exit_code == 1
    assert "the deployment cannot be published" in result.output


def test_chains_push_watch_with_environment_fails():
    """Test that --watch with --environment fails."""
    runner = CliRunner()

    mock_entrypoint_cls = Mock()
    mock_entrypoint_cls.meta_data.chain_name = "test_chain"
    mock_entrypoint_cls.display_name = "TestChain"

    with patch(
        "truss_chains.framework.ChainletImporter.import_target"
    ) as mock_importer:
        mock_importer.return_value.__enter__.return_value = mock_entrypoint_cls

        result = runner.invoke(
            truss_cli,
            [
                "chains",
                "push",
                "test_chain.py",
                "--watch",
                "--environment",
                "staging",
                "--remote",
                "test_remote",
            ],
        )

    assert result.exit_code == 1
    assert "Cannot use --watch with --environment" in result.output



import logging
import os
import threading
import time
from typing import Optional

from helpers.custom_types import Patch, PatchType
from helpers.errors import (
    InadmissiblePatch,
    PatchFailedRecoverable,
    PatchFailedUnrecoverable,
    UnsupportedPatch,
)
from helpers.inference_server_process_controller import InferenceServerProcessController
from helpers.truss_patch.model_container_patch_applier import ModelContainerPatchApplier

INFERENCE_SERVER_CHECK_INTERVAL_SECS = 10


class InferenceServerController:
    """Controls what code the inference server runs with.

    Currently, it only applies locks to various actions and mostly
    delegates to InferenceServerProcessController.
    """

    _process_controller: InferenceServerProcessController
    _patch_applier: ModelContainerPatchApplier
    _app_logger: logging.Logger
    _oversee_inference_server: bool
    _inf_env: dict

    def __init__(
        self,
        process_controller: InferenceServerProcessController,
        patch_applier: ModelContainerPatchApplier,
        app_logger: logging.Logger,
        oversee_inference_server: bool = True,
    ):
        self._lock = threading.Lock()
        self._process_controller = process_controller
        self._patch_applier = patch_applier
        self._current_running_hash = os.environ.get("HASH_TRUSS", None)
        self._app_logger = app_logger
        self._has_partially_applied_patch = False
        self._inf_env = (
            os.environ.copy()
        )  # this must be initialized before overseer is started
        if oversee_inference_server:
            self._inference_server_overseer_thread = threading.Thread(
                target=self._check_and_recover_inference_server
            )
            self._inference_server_overseer_thread.start()

    def apply_patch(self, patch_request):
        with self._lock:
            req_hash = patch_request["hash"]
            req_prev_hash = patch_request["prev_hash"]

            self._app_logger.debug(
                f"request hash: {req_hash}, request prev hash:"
                f" {req_prev_hash}, current hash: "
                f"{self._current_running_hash}"
            )

            if req_hash == self._current_running_hash:
                # We are in sync, ok to start running inference server now, if
                # it's not running.
                if not self._process_controller.inference_server_started():
                    self._process_controller.start(self._inf_env)
                self._app_logger.info("Request hash same as current hash, skipping.")
                return

            if req_prev_hash != self._current_running_hash:
                raise InadmissiblePatch(
                    f"Unable to apply patch: expected prev hash \
                    to be {self._current_running_hash} but found {req_prev_hash}"
                )

            try:
                patches = [
                    Patch.from_dict(patch_dict)
                    for patch_dict in patch_request["patches"]
                ]
            except (KeyError, ValueError) as exc:
                raise UnsupportedPatch(str(exc)) from exc

            self._process_controller.stop()
            patches.sort(key=_patch_sort_key_fn)
            try:
                patches_executed = 0
                for patch in patches:
                    self._patch_applier(patch, self._inf_env)
                    patches_executed += 1
            except Exception as exc:
                if patches_executed > 0:
                    # In this case we leave inference server stopped, to reflect
                    # the bad state; with partially applied patch all bets are off.
                    # Correct handling of this scenario is to fallback to full deploy.
                    self._has_partially_applied_patch = True
                    raise PatchFailedUnrecoverable(str(exc)) from exc
                # No patches executed, the very first patch failed. We
                # consider this safe to start inference server back up.
                # Theoretically, a single patch application may leave
                # side-effects, but that's not the case with currently
                # supported patches.
                self._process_controller.start(self._inf_env)
                raise PatchFailedRecoverable(str(exc)) from exc

            self._process_controller.start(self._inf_env)
            self._current_running_hash = req_hash

    def truss_hash(self) -> Optional[str]:
        return self._current_running_hash

    def restart(self):
        with self._lock:
            self._process_controller.stop()
            self._process_controller.start(self._inf_env)

    def start(self):
        # For now, start just does restart, this alias allows for better
        # readability in certain scenarios where the intention is to start.
        self.restart()

    def stop(self):
        with self._lock:
            self._process_controller.stop()

    def has_partially_applied_patch(self):
        with self._lock:
            return self._has_partially_applied_patch

    def _check_and_recover_inference_server(self):
        self._app_logger.info("Inference server overseer thread started")
        while not self._process_controller.is_inference_server_terminated():
            with self._lock:
                self._process_controller.check_and_recover_inference_server(
                    self._inf_env
                )
            time.sleep(INFERENCE_SERVER_CHECK_INTERVAL_SECS)


def _patch_sort_key_fn(patch: Patch) -> int:
    # System packages need to be applied before python requirements as they
    # might need them.
    # System packages or python packages are more likely to fail to apply then
    # model code patches, so it's better to apply model code patches last to
    # avoid ending up with partially applied patches which currently triggers
    # full deploy.

    PATCH_ORDER = [
        PatchType.SYSTEM_PACKAGE,
        PatchType.PYTHON_REQUIREMENT,
        PatchType.PACKAGE,
        PatchType.ENVIRONMENT_VARIABLE,
        PatchType.EXTERNAL_DATA,
        PatchType.CONFIG,
        PatchType.MODEL_CODE,
    ]
    if patch.type not in PATCH_ORDER:
        raise ValueError(f"Unexpected patch type {patch.type}")
    return PATCH_ORDER.index(patch.type)



import contextlib
import logging
import sys
import textwrap
from collections.abc import Generator, Mapping
from http import HTTPStatus
from types import TracebackType
from typing import Optional, Union

import fastapi
import pydantic
import starlette.responses
from fastapi.responses import JSONResponse

# See https://github.com/basetenlabs/baseten/blob/master/docs/Error-Propagation.md
_TRUSS_SERVER_SERVICE_ID = 4
_BASETEN_UNEXPECTED_ERROR = 500
_BASETEN_DOWNSTREAM_ERROR_CODE = 600
_BASETEN_CLIENT_ERROR_CODE = 700

MODEL_ERROR_MESSAGE = "Internal Server Error (in model/chainlet)."
WEBSOCKET_SERVER_ERROR_CODE = 1011


class ModelMissingError(Exception):
    def __init__(self, path):
        self.path = path

    def __str__(self):
        return self.path


class ModelNotReady(RuntimeError):
    def __init__(self, model_name: str, detail: Optional[str] = None):
        self.model_name = model_name
        self.error_msg = f"Model with name {self.model_name} is not ready."
        if detail:
            self.error_msg = self.error_msg + " " + detail

    def __str__(self):
        return self.error_msg


class InputParsingError(ValueError):
    pass


class UserCodeError(Exception):
    pass


class ModelMethodNotImplemented(Exception):
    pass


class ModelDefinitionError(TypeError):
    """When the user-defined truss model does not meet the contract."""


def _make_baseten_error_headers(error_code: int) -> Mapping[str, str]:
    return {
        "X-BASETEN-ERROR-SOURCE": f"{_TRUSS_SERVER_SERVICE_ID:02}",
        "X-BASETEN-ERROR-CODE": f"{error_code:03}",
    }


def add_error_headers_to_user_response(response: starlette.responses.Response) -> None:
    response.headers.update(_make_baseten_error_headers(_BASETEN_CLIENT_ERROR_CODE))


def _make_baseten_response(
    http_status: int, info: Union[str, Exception], baseten_error_code: int
) -> fastapi.Response:
    msg = str(info) if isinstance(info, Exception) else info
    return JSONResponse(
        status_code=http_status,
        content={"error": msg},
        headers=_make_baseten_error_headers(baseten_error_code),
    )


async def exception_handler(_: fastapi.Request, exc: Exception) -> fastapi.Response:
    if isinstance(exc, ModelMissingError):
        return _make_baseten_response(
            HTTPStatus.NOT_FOUND.value, exc, _BASETEN_DOWNSTREAM_ERROR_CODE
        )
    if isinstance(exc, ModelNotReady):
        return _make_baseten_response(
            HTTPStatus.SERVICE_UNAVAILABLE.value, exc, _BASETEN_DOWNSTREAM_ERROR_CODE
        )
    if isinstance(exc, InputParsingError):
        return _make_baseten_response(
            HTTPStatus.BAD_REQUEST.value, exc, _BASETEN_CLIENT_ERROR_CODE
        )
    if isinstance(exc, ModelDefinitionError):
        return _make_baseten_response(
            HTTPStatus.PRECONDITION_FAILED.value,
            f"{type(exc).__name__}: {str(exc)}",
            _BASETEN_DOWNSTREAM_ERROR_CODE,
        )
    if isinstance(exc, UserCodeError):
        return _make_baseten_response(
            HTTPStatus.INTERNAL_SERVER_ERROR.value,
            MODEL_ERROR_MESSAGE,
            _BASETEN_DOWNSTREAM_ERROR_CODE,
        )
    if isinstance(exc, ModelMethodNotImplemented):
        return _make_baseten_response(
            HTTPStatus.NOT_FOUND.value, exc, _BASETEN_CLIENT_ERROR_CODE
        )
    if isinstance(exc, fastapi.HTTPException):
        # This is a pass through, but additionally adds our custom error headers.
        return _make_baseten_response(
            exc.status_code, exc.detail, _BASETEN_DOWNSTREAM_ERROR_CODE
        )
    # Fallback case.
    return _make_baseten_response(
        HTTPStatus.INTERNAL_SERVER_ERROR.value,
        f"Unhandled exception: {type(exc).__name__}: {str(exc)}",
        _BASETEN_UNEXPECTED_ERROR,
    )


HANDLED_EXCEPTIONS = {
    ModelMissingError,
    ModelNotReady,
    NotImplementedError,
    InputParsingError,
    UserCodeError,
    ModelDefinitionError,
    fastapi.HTTPException,
    ModelMethodNotImplemented,
}


def filter_traceback(
    model_file_name: str,
) -> Union[
    tuple[type[BaseException], BaseException, TracebackType], tuple[None, None, None]
]:
    exc_type, exc_value, tb = sys.exc_info()
    if tb is None:
        return exc_type, exc_value, tb  # type: ignore[return-value]

    # Store the last occurrence of the traceback matching the condition
    last_matching_tb: Optional[TracebackType] = None
    current_tb: Optional[TracebackType] = tb

    while current_tb is not None:
        filename = current_tb.tb_frame.f_code.co_filename
        if filename.endswith(model_file_name):
            last_matching_tb = current_tb
        current_tb = current_tb.tb_next

    # If a match was found, truncate the traceback at the last occurrence
    if last_matching_tb is not None:
        return exc_type, exc_value, last_matching_tb  # type: ignore[return-value]

    # If `model_file_name` not found, return the original exception info
    return exc_type, exc_value, tb  # type: ignore[return-value]


@contextlib.contextmanager
def intercept_exceptions(
    logger: logging.Logger, model_file_name: str
) -> Generator[None, None, None]:
    try:
        yield
    # Note that logger.error logs the stacktrace, such that the user can
    # debug this error from the logs.
    except fastapi.HTTPException as e:
        # TODO: we try to avoid any dependency of the truss server on chains, but for
        #  the purpose of getting readable chained-stack traces in the server logs,
        #  we have to add a special-case here.
        if "user_stack_trace" in e.detail:
            try:
                try:
                    from truss_chains import public_types

                    error_cls = public_types.RemoteErrorDetail
                except ImportError:
                    # For pre 0.9.67 usage.
                    from truss_chains import definitions  # type: ignore[attr-defined]

                    error_cls = definitions.RemoteErrorDetail

                chains_error = error_cls.model_validate(e.detail)
                # The formatted error contains a (potentially chained) stack trace
                # with all framework code removed, see
                # truss_chains/remote_chainlet/utils.py::response_raise_errors.
                logger.error(f"Chainlet raised Exception:\n{chains_error.format()}")
            except:  # If we cannot import chains or parse the error.
                logger.error(
                    "Model raised HTTPException",
                    exc_info=filter_traceback(model_file_name),
                )
                raise
            # If error was extracted successfully, the customized stack trace is
            # already printed above, so we raise with a clear traceback.
            e.__traceback__ = None
            raise e from None

        logger.error(
            "Model raised HTTPException", exc_info=filter_traceback(model_file_name)
        )
        raise
    except Exception as exc:
        logger.error(MODEL_ERROR_MESSAGE, exc_info=filter_traceback(model_file_name))
        raise UserCodeError(str(exc))


def _loc_to_dot_sep(loc: tuple[Union[str, int], ...]) -> str:
    # From https://docs.pydantic.dev/latest/errors/errors/#customize-error-messages.
    # Chained field access is stylized with `.`-notation (corresponding to str parts)
    # and array indexing is stylized with `[i]`-notation (corresponding to int parts).
    # E.g. ('items', 1, 'value') -> items[1].value.
    parts = []
    for i, x in enumerate(loc):
        if isinstance(x, str):
            if i > 0:
                parts.append(".")
            parts.append(x)
        elif isinstance(x, int):
            parts.append(f"[{x}]")
        else:
            raise TypeError(f"Unexpected type: {x}.")
    return "".join(parts)


def format_pydantic_validation_error(exc: pydantic.ValidationError) -> str:
    if exc.error_count() == 1:
        parts = ["Input Parsing Error:"]
    else:
        parts = ["Input Parsing Errors:"]

    try:
        for error in exc.errors(include_url=False, include_input=False):
            loc = _loc_to_dot_sep(error["loc"])
            if error["msg"]:
                msg = error["msg"]
            else:
                error_no_loc = dict(error)
                error_no_loc.pop("loc")
                msg = str(error_no_loc)
            parts.append(textwrap.indent(f"`{loc}`: {msg}.", "  "))
    except:  # noqa: E722
        # Fallback in case any of the fields cannot be processed as expected.
        return f"Input Parsing Error, {str(exc)}"

    return "\n".join(parts)



import base64
import json
import os
import tarfile
import tempfile
from dataclasses import dataclass
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Callable, Dict, Optional, Tuple

import click
import requests
import rich
from InquirerPy import inquirer
from rich.text import Text

from truss.cli.train import common, deploy_checkpoints
from truss.cli.train.metrics_watcher import MetricsWatcher
from truss.cli.train.types import (
    DeployCheckpointArgs,
    DeployCheckpointsConfigComplete,
    DeploySuccessResult,
)
from truss.cli.utils import common as cli_common
from truss.cli.utils.output import console
from truss.remote.baseten.remote import BasetenRemote
from truss_train import loader
from truss_train.definitions import DeployCheckpointsConfig

ACTIVE_JOB_STATUSES = [
    "TRAINING_JOB_RUNNING",
    "TRAINING_JOB_CREATED",
    "TRAINING_JOB_DEPLOYING",
]


def _get_job_by_job_id(remote_provider: BasetenRemote, job_id: str) -> dict:
    jobs = remote_provider.api.search_training_jobs(job_id=job_id)
    if not jobs:
        raise RuntimeError(f"No training job found with ID: {job_id}")
    return jobs[0]


def _get_latest_job(remote_provider: BasetenRemote) -> dict:
    jobs = remote_provider.api.search_training_jobs(
        order_by=[{"field": "created_at", "order": "desc"}]
    )
    if not jobs:
        raise click.ClickException("No training jobs found. Please start a job first.")
    return jobs[0]


def get_args_for_stop(
    remote_provider: BasetenRemote, project_id: Optional[str], job_id: Optional[str]
) -> Tuple[str, str]:
    if not project_id or not job_id:
        # get all running jobs
        job = _get_active_job(remote_provider, project_id, job_id)
        project_id_for_job = job["training_project"]["id"]
        job_id_to_stop = job["id"]
        # check if the user wants to stop the inferred running job
        if not job_id:
            confirm = inquirer.confirm(
                message=f"Are you sure you want to stop training job {job_id_to_stop}?",
                default=False,
            ).execute()
            if not confirm:
                raise click.UsageError("Training job not stopped.")
        return project_id_for_job, job_id_to_stop

    return project_id, job_id


def _get_active_job(
    remote_provider: BasetenRemote, project_id: Optional[str], job_id: Optional[str]
) -> dict:
    jobs = remote_provider.api.search_training_jobs(
        statuses=ACTIVE_JOB_STATUSES, project_id=project_id, job_id=job_id
    )
    if not jobs:
        raise click.UsageError("No running jobs found.")
    if len(jobs) > 1:
        display_training_jobs(
            jobs, remote_provider.remote_url, title="Active Training Jobs"
        )
        raise click.UsageError("Multiple active jobs found. Please specify a job id.")
    return jobs[0]


@dataclass
class DisplayTableColumn:
    name: str
    style: dict
    accessor: Callable[[dict], str]


def display_training_jobs(
    jobs, remote_url: str, checkpoints_by_job_id=None, title="Training Job Details"
):
    checkpoints_by_job_id = checkpoints_by_job_id or {}
    console.print(title, style="bold magenta")
    for job in jobs:
        display_training_job(job, remote_url, checkpoints_by_job_id.get(job["id"], []))


def recreate_training_job(
    remote_provider: BasetenRemote, job_id: Optional[str] = None
) -> Dict[str, Any]:
    job: dict
    if job_id:
        job = _get_job_by_job_id(remote_provider, job_id)
    else:
        job = _get_latest_job(remote_provider)
        job_id = job["id"]
        confirm = inquirer.confirm(
            message=f"Recreate training job from most recent job {job_id}?",
            default=False,
        ).execute()

        if not confirm:
            raise click.UsageError("Training job not recreated.")

    project_id = job["training_project"]["id"]
    job_id = job["id"]
    job_resp = remote_provider.api.recreate_training_job(project_id, job_id)  # type: ignore
    return job_resp


def display_training_projects(projects: list[dict], remote_url: str) -> None:
    table = rich.table.Table(
        show_header=True,
        header_style="bold magenta",
        title="Training Projects",
        box=rich.table.box.ROUNDED,
        border_style="blue",
    )
    table.add_column("Project ID", style="cyan")
    table.add_column("Name")
    table.add_column("Created")
    table.add_column("Last Modified")
    table.add_column("Latest Job ID", style="bold yellow")
    table.add_column("Latest Job Status", style="bold yellow")
    table.add_column("Status Page", style="bold yellow")

    # most recent projects at bottom of terminal
    for project in projects[::-1]:
        latest_job = project.get("latest_job") or {}
        if latest_job_id := latest_job.get("id", ""):
            latest_job_link = cli_common.format_link(
                status_page_url(remote_url, project["id"], latest_job_id), "link"
            )
        else:
            latest_job_link = ""
        table.add_row(
            project["id"],
            project["name"],
            cli_common.format_localized_time(project["created_at"]),
            cli_common.format_localized_time(project["updated_at"]),
            latest_job_id,
            latest_job.get("current_status", ""),
            latest_job_link,
        )

    console.print(table)


def view_training_details(
    remote_provider: BasetenRemote, project_id: Optional[str], job_id: Optional[str]
):
    """
    view_training_details shows a list of jobs that meet the provided project_id and job_id filters.

     If no filters are provided, the command will show a list of all training projects and a list of active jobs.
    """
    if job_id or project_id:
        jobs_response = remote_provider.api.search_training_jobs(
            project_id=project_id,
            job_id=job_id,
            order_by=[{"field": "created_at", "order": "asc"}],
        )
        if len(jobs_response) == 0:
            raise click.UsageError("No training jobs found")
        if len(jobs_response) == 1:
            training_job = jobs_response[0]
            checkpoints = remote_provider.api.list_training_job_checkpoints(
                training_job["training_project"]["id"], training_job["id"]
            )
            display_training_job(
                training_job, remote_provider.remote_url, checkpoints["checkpoints"]
            )
        else:
            display_training_jobs(jobs_response, remote_provider.remote_url)
    else:
        projects = remote_provider.api.list_training_projects()
        display_training_projects(projects, remote_provider.remote_url)
        active_jobs = remote_provider.api.search_training_jobs(
            statuses=ACTIVE_JOB_STATUSES
        )
        if active_jobs:
            display_training_jobs(
                active_jobs, remote_provider.remote_url, title="Active Training Jobs"
            )
        else:
            console.print("No active training jobs.", style="yellow")


def stop_all_jobs(remote_provider: BasetenRemote, project_id: Optional[str]):
    active_jobs = remote_provider.api.search_training_jobs(
        project_id=project_id, statuses=ACTIVE_JOB_STATUSES
    )
    if not active_jobs:
        console.print("No active jobs found.", style="yellow")
        return
    confirm = inquirer.confirm(
        message=f"Are you sure you want to stop {len(active_jobs)} active jobs?",
        default=False,
    ).execute()
    if not confirm:
        raise click.UsageError("Training jobs not stopped.")
    for job in active_jobs:
        remote_provider.api.stop_training_job(job["training_project"]["id"], job["id"])
    console.print("Training jobs stopped successfully.", style="green")


def view_training_job_metrics(
    remote_provider: BasetenRemote, project_id: Optional[str], job_id: Optional[str]
):
    """
    view_training_job_metrics shows a list of metrics for a training job.
    """
    project_id, job_id = common.get_most_recent_job(remote_provider, project_id, job_id)
    metrics_display = MetricsWatcher(remote_provider.api, project_id, job_id)
    metrics_display.watch()


def create_model_version_from_inference_template(
    remote_provider: BasetenRemote, args: DeployCheckpointArgs
) -> DeploySuccessResult:
    if not args.deploy_config_path:
        return deploy_checkpoints.create_model_version_from_inference_template(
            remote_provider,
            DeployCheckpointsConfig(),
            args.project_id,
            args.job_id,
            args.dry_run,
        )
    # User provided a checkpoint deploy config file
    with loader.import_deploy_checkpoints_config(
        Path(args.deploy_config_path)
    ) as checkpoint_deploy:
        return deploy_checkpoints.create_model_version_from_inference_template(
            remote_provider,
            checkpoint_deploy,
            args.project_id,
            args.job_id,
            args.dry_run,
        )


def _get_checkpoint_names(
    checkpoint_deploy_config: DeployCheckpointsConfigComplete,
) -> list[str]:
    return [
        checkpoint.checkpoint_name
        for checkpoint in checkpoint_deploy_config.checkpoint_details.checkpoints
    ]


def print_deploy_checkpoints_success_message(
    checkpoint_deploy_config: DeployCheckpointsConfigComplete,
):
    checkpoint_names = _get_checkpoint_names(checkpoint_deploy_config)
    console.print(
        Text("\nTo run the model"),
        Text("ensure your `model` parameter is set to one of"),
        Text(
            f"{[checkpoint_name for checkpoint_name in checkpoint_names]}",
            style="magenta",
        ),
        Text("in your request. An example request body might look like this:"),
        Text(
            f'\n{{"model": "{checkpoint_names[0]}", "messages": [...]}}', style="green"
        ),
    )


def display_training_job(
    job: dict, remote_url: str, checkpoints: Optional[list[dict]] = None
):
    table = rich.table.Table(
        show_header=False,
        title=f"Training Job: {job['id']}",
        box=rich.table.box.ROUNDED,
        border_style="blue",
        title_style="bold magenta",
    )
    table.add_column("Field", style="bold cyan")
    table.add_column("Value")

    # Basic job details
    table.add_row("Job Name", job["name"])
    table.add_row("Job ID", job["id"])
    table.add_row("Project ID", job["training_project"]["id"])
    table.add_row("Project Name", job["training_project"]["name"])
    table.add_row("Status", job["current_status"])
    table.add_row("Instance Type", job["instance_type"]["name"])
    table.add_row("Created", cli_common.format_localized_time(job["created_at"]))
    table.add_row("Last Modified", cli_common.format_localized_time(job["updated_at"]))
    table.add_row(
        "Job Page",
        cli_common.format_link(
            status_page_url(remote_url, job["training_project"]["id"], job["id"]),
            "link",
        ),
    )

    # Add error message if present
    if job.get("error_message"):
        table.add_row("Error Message", Text(job["error_message"], style="red"))

    # Add checkpoints if present
    if checkpoints:
        checkpoint_text = ", ".join(
            [checkpoint["checkpoint_id"] for checkpoint in checkpoints]
        )
        table.add_row("Checkpoints", checkpoint_text)

    console.print(table)


def _generate_job_artifact_name(project_name: str, job_id: str) -> str:
    return f"{project_name}_{job_id}"


def download_training_job_data(
    remote_provider: BasetenRemote,
    job_id: str,
    target_directory: Optional[str],
    unzip: bool,
) -> Path:
    output_dir = Path(target_directory).resolve() if target_directory else Path.cwd()
    output_dir.mkdir(parents=True, exist_ok=True)

    job = _get_job_by_job_id(remote_provider, job_id)

    project = job["training_project"]
    project_id = project["id"]
    project_name = project["name"]

    artifact_base_name = _generate_job_artifact_name(project_name, job_id)
    file_name = f"{artifact_base_name}.tgz"
    target_path = output_dir / file_name

    presigned_url = remote_provider.api.get_training_job_presigned_url(
        project_id=project_id, job_id=job_id
    )
    content = remote_provider.api.get_from_presigned_url(presigned_url)

    if unzip:
        with tempfile.NamedTemporaryFile() as temp_file:
            temp_path = Path(temp_file.name)
            temp_path.write_bytes(content)

            unzip_dir = output_dir / artifact_base_name
            unzip_dir = Path(str(unzip_dir).replace(" ", "-"))
            if unzip_dir.exists():
                raise click.ClickException(
                    f"Directory '{unzip_dir}' already exists. "
                    "Please remove it or specify a different target directory."
                )

            unzip_dir.mkdir(parents=True, exist_ok=True)
            with tarfile.open(temp_path, "r:*") as tar:
                tar.extractall(path=unzip_dir)

            return unzip_dir
    else:
        target_path = Path(str(target_path).replace(" ", "-"))
        target_path.write_bytes(content)
        return target_path


def download_checkpoint_artifacts(
    remote_provider: BasetenRemote, job_id: Optional[str]
) -> Path:
    output_dir = Path.cwd()
    job: dict

    if job_id:
        job = _get_job_by_job_id(remote_provider, job_id)
    else:
        job = _get_latest_job(remote_provider)

    job_id = job["id"]
    project = job["training_project"]
    project_id = project["id"]
    project_name = project["name"]

    checkpoint_artifacts = (
        remote_provider.api.get_training_job_checkpoint_presigned_url(
            project_id=project_id,
            job_id=job_id,  # type: ignore
            page_size=1000,
        )
    )

    if not checkpoint_artifacts:
        raise click.ClickException("No checkpoints found for this training job.")

    output = {
        "timestamp": datetime.now(timezone.utc).isoformat(),
        "job": job,
        "checkpoint_artifacts": checkpoint_artifacts,
    }

    urls_file = (
        output_dir / f"{project_name.replace(' ', '-')}_{job_id}_checkpoints.json"
    )
    with open(urls_file, "w") as f:
        json.dump(output, f, indent=2)

    return urls_file


def status_page_url(remote_url: str, project_id: str, training_job_id: str) -> str:
    return f"{remote_url}/training/{project_id}/logs/{training_job_id}"


def _get_all_train_init_example_options(
    repo_id: str = "ml-cookbook",
    examples_subdir: str = "examples",
    token: Optional[str] = None,
) -> list[str]:
    """
    Retrieve a list of all example options from the ml-cookbook repository to
    copy locally for training initialization. This method generates a list
    of examples and URL paths to show the user for selection.
    """
    headers = {}
    if token:
        headers["Authorization"] = f"token {token}"

    url = (
        f"https://api.github.com/repos/basetenlabs/{repo_id}/contents/{examples_subdir}"
    )
    try:
        response = requests.get(url, headers=headers)
        response.raise_for_status()

        items = response.json()
        if not isinstance(items, list):
            items = [items]
        items = [item["name"] for item in items if item["type"] == "dir"]
        return items

    except requests.exceptions.RequestException as e:
        click.echo(
            f"Error exploring directory: {e}. Please file an issue at https://github.com/basetenlabs/truss/issues"
        )
        return []


def _get_train_init_example_info(
    repo_id: str = "ml-cookbook",
    examples_subdir: str = "examples",
    training_subdir: str = "training",
    example_name: Optional[str] = None,
    token: Optional[str] = None,
) -> list[Dict[str, str]]:
    """
    Retrieve directory download links for the example from the ml-cookbook repository to
    copy locally for training initialization.
    """
    headers = {}
    if token:
        headers["Authorization"] = f"token {token}"

    url = f"https://api.github.com/repos/basetenlabs/{repo_id}/contents/{examples_subdir}/{example_name}"

    console.print(f"Attempting to retrieve example info from: {url}")
    try:
        response = requests.get(url, headers=headers)
        response.raise_for_status()

        items = response.json()
        if not isinstance(items, list):
            items = [items]
        # return only training subdirectory info for example
        items = [item for item in items if item["name"] == training_subdir]
        return items

    except requests.exceptions.HTTPError as e:
        if response.status_code == 404:
            # example_name does not exist, return empty list
            return []
        else:
            # Other HTTP errors
            click.echo(
                f"Error exploring directory: {e}. Please file an issue at https://github.com/basetenlabs/truss/issues"
            )
            return []
    except requests.exceptions.RequestException as e:
        # Network or other request errors
        click.echo(
            f"Error exploring directory: {e}. Please file an issue at https://github.com/basetenlabs/truss/issues"
        )
        return []


def download_git_directory(
    git_api_url: str, local_dir: str, token: Optional[str] = None
):
    """
    Recursively download directory contents from git api url.
    Special handling for 'training' directory: downloads its contents directly
    to local_dir without creating a 'training' subdirectory.
    Args:
        git_api_url (str): Example format "https://api.github.com/repos/basetenlabs/ml-cookbook/contents/examples/llama-finetune-8b-lora?ref=main"
        local_dir(str): Local directory to download this directory to
    """
    headers = {}
    if token:
        headers["Authorization"] = f"token {token}"
    try:
        response = requests.get(git_api_url, headers=headers)
        response.raise_for_status()
        items = response.json()

        # Handle single file case
        if not isinstance(items, list):
            items = [items]

        # Create local directory
        print(f"Creating directory {local_dir}")
        os.makedirs(local_dir, exist_ok=True)

        # Check if there's a 'training' directory in the items
        training_dir = None
        other_items = []

        for item in items:
            if item["name"] == "training" and item["type"] == "dir":
                training_dir = item
            else:
                other_items.append(item)

        # If training directory exists, download its contents directly to local_dir
        if training_dir:
            print(
                f"📁 Found training directory, downloading its contents to {local_dir}"
            )
            return download_git_directory(training_dir["url"], local_dir)

        # If no training directory, download all files normally
        for item in other_items:
            item_name = item["name"]
            local_item_path = os.path.join(local_dir, item_name)

            if item["type"] == "file":
                print(f"📄 Downloading {item_name}")
                if item.get("download_url"):
                    # Download file directly
                    file_response = requests.get(item["download_url"])
                    file_response.raise_for_status()
                    with open(local_item_path, "wb") as f:
                        f.write(file_response.content)
                elif item.get("content"):
                    # Decode base64 content (for small files)
                    try:
                        content = base64.b64decode(item["content"])
                        with open(local_item_path, "wb") as f:
                            f.write(content)
                    except Exception as e:
                        print(f"⚠️ Could not decode {item_name}: {e}")
            elif item["type"] == "dir":
                print(f"📁 Entering directory {item_name}")
                # Use the API URL from the response for subdirectories
                download_git_directory(item["url"], local_item_path)
        return True
    except Exception as e:
        print(f"Error processing response: {e}")
        return False


def fetch_project_by_name_or_id(
    remote_provider: BasetenRemote, project_identifier: str
) -> dict:
    """Fetch a training project by name or ID.

    Args:
        remote_provider: The remote provider instance
        project_identifier: Either a project ID or project name

    Returns:
        The project object as a dictionary

    Raises:
        click.ClickException: If the project is not found
    """
    try:
        projects = remote_provider.api.list_training_projects()
        projects_by_name = {project.get("name"): project for project in projects}
        projects_by_id = {project.get("id"): project for project in projects}
        if project_identifier in projects_by_id:
            return projects_by_id[project_identifier]
        if project_identifier in projects_by_name:
            return projects_by_name[project_identifier]
        valid_project_ids_and_names = ", ".join(
            [f"{project.get('id')} ({project.get('name')})" for project in projects]
        )
        raise click.ClickException(
            f"Project '{project_identifier}' not found. Valid project IDs and names: {valid_project_ids_and_names}"
        )
    except click.ClickException:
        raise
    except Exception as e:
        raise click.ClickException(f"Error fetching project: {str(e)}")


def view_cache_summary_by_project(
    remote_provider: BasetenRemote,
    project_identifier: str,
    sort_by: Optional[str] = None,
    order: Optional[str] = None,
    output_format: Optional[str] = None,
):
    """View cache summary for a training project by ID or name."""
    from truss.cli.train.cache import (
        OUTPUT_FORMAT_CLI_TABLE,
        SORT_BY_FILEPATH,
        SORT_ORDER_ASC,
        view_cache_summary,
    )

    # Use constants for defaults if not provided
    if sort_by is None:
        sort_by = SORT_BY_FILEPATH
    if order is None:
        order = SORT_ORDER_ASC
    if output_format is None:
        output_format = OUTPUT_FORMAT_CLI_TABLE

    project = fetch_project_by_name_or_id(remote_provider, project_identifier)
    view_cache_summary(remote_provider, project["id"], sort_by, order, output_format)



from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from typing import Optional
from unittest.mock import patch

import pytest

from truss.cli.train_commands import _format_time_until_expiry


def _utc_iso(dt: datetime) -> str:
    return dt.strftime("%Y-%m-%dT%H:%M:%SZ")


def _fixed_now(target: datetime):
    """Return a patcher that freezes datetime.now to *target*."""
    return patch(
        "truss.cli.train_commands.datetime",
        wraps=datetime,
        **{"now.return_value": target},
    )


NOW = datetime(2026, 2, 17, 12, 0, 0, tzinfo=timezone.utc)


@dataclass
class FormatTimeTillExpiryTestCase:
    desc: str
    input: Optional[str]
    expected: str
    freeze_now: bool = True


CASES = [
    FormatTimeTillExpiryTestCase(
        desc="empty string returns empty", input="", expected="", freeze_now=False
    ),
    FormatTimeTillExpiryTestCase(
        desc="None returns empty", input=None, expected="", freeze_now=False
    ),
    FormatTimeTillExpiryTestCase(
        desc="hours and minutes",
        input=_utc_iso(NOW + timedelta(hours=2, minutes=30)),
        expected="2h 30m",
    ),
    FormatTimeTillExpiryTestCase(
        desc="hours only (zero minutes)",
        input=_utc_iso(NOW + timedelta(hours=4)),
        expected="4h 0m",
    ),
    FormatTimeTillExpiryTestCase(
        desc="minutes only", input=_utc_iso(NOW + timedelta(minutes=45)), expected="45m"
    ),
    FormatTimeTillExpiryTestCase(
        desc="less than one minute remaining",
        input=_utc_iso(NOW + timedelta(seconds=30)),
        expected="< 1m",
    ),
    FormatTimeTillExpiryTestCase(
        desc="already expired (past)",
        input=_utc_iso(NOW - timedelta(minutes=5)),
        expected="Expired",
    ),
    FormatTimeTillExpiryTestCase(
        desc="exactly expired (boundary)", input=_utc_iso(NOW), expected="Expired"
    ),
    FormatTimeTillExpiryTestCase(
        desc="invalid timestamp returns original",
        input="not-a-date",
        expected="not-a-date",
        freeze_now=False,
    ),
    FormatTimeTillExpiryTestCase(
        desc="ISO with Z suffix", input="2026-02-17T13:15:00Z", expected="1h 15m"
    ),
    FormatTimeTillExpiryTestCase(
        desc="ISO with +00:00 offset",
        input="2026-02-17T13:00:00+00:00",
        expected="1h 0m",
    ),
]


@pytest.mark.parametrize("case", CASES, ids=[c.desc for c in CASES])
def test_format_time_until_expiry(case: FormatTimeTillExpiryTestCase):
    if case.freeze_now:
        with _fixed_now(NOW):
            assert _format_time_until_expiry(case.input) == case.expected
    else:
        assert _format_time_until_expiry(case.input) == case.expected
